/*
 * Frost - use C++ multi methods as if they were a native feature
 * Copyright (C) 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/stdexcpt.hh>
#include <uds/stl_util.hh>
#include <uds/str_util.hh>
#include "class.hh"
#include "progb.hh"
#include "table.hh"
#include "vtable.hh"

using namespace uds;


// Class::vinfo implementation

Class::vinfo::vinfo( size_t to, int d ) :
	t_offs( to ), dist( d )
{
}


// Class implementation

namespace
{

char
echar( char c )
{
	if ( isupper( c )
		 || islower( c )
		 || isdigit( c )
		 || c == '_' )
		return c;

	return '_';
}

}


Class::Class( const string& n, size_t sz, bool b ) :
	name( n ), size( sz ), bit( b ), tainted( false )
{
	_frost_Log( LOG_DEBUG + 1, ( b ? "new builtin " : "new class " ) + n );
	
	if ( n.empty() )
		_uds_abort;
}

string
Class::EName() const
{
	string ret;
	ret.resize( name.size() );

	for ( size_t i = 0; i < name.size(); ++i )
		ret[i] = echar( name[i] );

	ret.insert( 0, StringVar( ret.size() ));

	return ret;
}

void
Class::SetSize( size_t sz )
{
	if ( size )
		_uds_abort;

	size = sz;
}

bool
Class::Match( Class& h, int& m )
{
	if ( Name() == h.Name() )
	{
		m = 0;
		return true;
	}
	
	vinfo* v = MapQuery( bases, &h );
	if ( v )
	{
		m = v->dist;
		return true;
	}

	v = MapQuery( h.bases, this );
	if ( v )
	{
		m = -v->dist;
		return true;
	}

	return false;
}

size_t
Class::TableOffset( Class& cl ) const
{
	return MapFindC< internal_error >( bases, &cl ).t_offs;
}

void
Class::SetTableOffset( Class& cl, size_t o, int d )
{
	_frost_Log( LOG_DEBUG + 2, "vtable: " + name + " -> " + cl.name
				+ " offset " + StringVar( o ) + " distance "
				+ StringVar( d ) );

	if ( &cl != this )
	{
		cl.derivs.insert( this );

		vinfo_map::iterator i = bases.find( &cl );
		if ( i != bases.end() )
		{
			if ( d < i->second.dist )
				i->second = vinfo( o, d );

			return;
		}
	}

	MapInsert< internal_error >( bases, &cl, vinfo( o, d ) );
}

void
Class::AddVBase( Class& cl )
{
	_frost_Log( LOG_DEBUG + 2, "vbase: " + name + " -> " + cl.name );

	for ( vinfo_map::const_iterator i = cl.bases.begin();
		  i != cl.bases.end(); ++i )
		vbases.insert( i->first );
}

const VTable&
Class::Vtbl() const
{
	if ( !vtbl )
		throw error( "virtual table information for class " + name + "\n"
					 "is not available." );

	return *vtbl;
}

void
Class::SetVtbl( VTable& vt )
{
	if ( &vt.Owner() != this )
		_uds_abort;

	vtbl = &vt;
}

Table&
Class::AddFTable( const string& n, TableGrp& tg, size_t fvo )
{
	tables.push_back( new Table( n, tg, *this, fvo ) );
	return *tables.back();
}

Table&
Class::MainFTable()
{
	if ( tables.empty() )
		_uds_abort;

	return *tables[0];
}

void
Class::DTableSet( table_set& ts )
{
	if ( tables.empty() )
		throw error( "class " + Name() + " has no virtual table" );

	ts.clear();
	
	tables[0]->DTableSet( ts );
}

void
Class::Taint()
{
	_frost_Log( LOG_WARNING, "class " + name + " tainted" );
	
	tainted = true;
}

Class::~Class()
{
	for ( size_t i = 0; i < tables.size(); ++i )
		delete tables[i];
}


// misc. implementation

ostream&
operator << ( ostream& out, const Class& cl )
{
	if ( cl.BIType() )
		return out;
	
	out<<"(\""<<cl.name<<"\" "<<cl.size<<' '<<( cl.tainted ? "#t" : "#f" )
	   <<" (";

	for ( size_t i = 0; i < cl.tables.size(); ++i )
	{
		if ( i )
			out<<' ';

		out<<*cl.tables[i];
	}

	out<<") (";

	bool f = true;
	for ( Class::vinfo_map::const_iterator i = cl.bases.begin();
		  i != cl.bases.end(); ++i )
	{
		if ( f )
			f = false;
		else
			out<<' ';

		out<<"(\""<<i->first->Name()<<"\" "<<i->second.t_offs
		   <<' '<<i->second.dist<<')';
	}

	out<<") (";

	f = true;
	for ( Class::class_set::const_iterator i = cl.vbases.begin();
		  i != cl.vbases.end(); ++i )
	{
		if ( f )
			f = false;
		else
			out<<' ';

		out<<'"'<<( *i )->Name()<<'"';
	}

	return out<<"))";
}
