/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/io_util.hh>
#include <uds/procstream.hh>
#include <uds/stdexcpt.hh>
#include <uds/str_util.hh>
#include "exception.hh"
#include "function.hh"
#include "progb.hh"


// Function implementation

namespace
{

void
skip( const string& dn, size_t& o, char s )
{
	while ( true )
	{
		char c = dn.at( o++ );

		if ( c == s )
			return;
		else if ( c == '(' )
			skip( dn, o, ')' );
		else if ( c == '<' )
			skip( dn, o, '>' );
	}
}

}


Function::Function( const string& en ) :
	ename( en ), ismm( false ), isc( false ), isp( false )
{
	_frost_Log( LOG_DEBUG + 1, "new function " + en );
}

string
Function::DemName( bool rm ) const
{
	static ProcStream ps( "c++filt" );

	ps<<ename<<endl;
	string dn = ReadLine( ps );

	if ( args.empty() )
		_uds_abort;

	if ( args[0].VArg() )
		dn.insert( 0, "virtual " );

	if ( args[0].SArg() )
		dn.insert( 0, "static " );

	if ( rm )
	{
		size_t so = dn.find( SYM_PREFIX );
		
		if ( so != string::npos )
		{
			size_t len = strlen( SYM_PREFIX );
			while ( dn.at( so + len++ ) != '_' );

			dn.replace( so, len, "" );
		}
	}

	size_t o = 0;
	while ( dn.at( o++ ) != '(' );

	if ( args.size() > 1 )
	{
		for ( size_t i = 1; ; )
		{
			if ( args[i].VArg() )
			{
				dn.insert( o, "virtual " );
				o += 8;
			}

			if ( ++i == args.size() )
				break;

			skip( dn, o, ',' );
			++o;
		}
	}
	
	return dn;
}

void
Function::SetFName( const string& fn )
{
	if ( fname.size() )
		_uds_abort;

	fname = fn;
}

void
Function::AddArg( const Type& tp )
{
	_frost_Log( LOG_DEBUG + 2, "function arg " + StringVar( tp ) );

	if ( args.size() && tp.Size() == 0 )
		_uds_abort;
	
	args.push_back( tp );
	ismm |= tp.VArg();
}

Function::cl_vec
Function::VAClassVec() const
{
	if ( !ismm )
		_uds_abort;
	
	cl_vec ret;

	for ( size_t i = 0; i < args.size(); ++i )
	{
		if ( !args[i].VArg() )
			continue;

		ret.push_back( args[i].GetClass() );
	}

	return ret;
}

Function::so_vec
Function::VASOffsVec() const
{
	if ( !ismm )
		_uds_abort;

	so_vec	ret;

	for ( size_t i = 0, offs = sizeof( void* ); i < args.size(); ++i )
	{
		if ( args[i].VArg() )
			ret.push_back( offs );

		offs += args[i].Size();
	}

	return ret;
}

bool
Function::IsStatic() const
{
	if ( args.empty() )
		_uds_abort;

	return args[0].SArg();
}

void
Function::SanityCheck() const
{
	string err;
	
	for ( size_t i = 1; i < args.size(); ++i )
	{
		try
		{
			args[i].SanityCheck();
		}
		catch ( InvVArg& e )
		{
			err += "arg " + StringVar( i ) + ": virtual argument must be a "
				"pointer or a reference to a class\n";
		}
	}

	if ( !err.empty() )
	{
		err.insert( 0, "\n" );
		err.insert( 0, DemName( true ) );

		throw InvVArg( err );
	}
}

int
Function::MMG_Compare( const Function& fn ) const
{
	if ( fname != fn.fname )
		_uds_abort;

	if ( args.size() != fn.args.size() || isc != fn.isc )
		return -1;

	int ret = 1;

	for ( size_t i = 0; i < args.size(); ++i )
	{
		int m;
		if ( !args[i].Compare( fn.args[i], m, !( i == 0 && IsStatic() )))
			return -1;

		if ( m > 0 )
			ret = 0;
	}

	return ret;
}

Function::~Function()
{
}


// misc. implementation

ostream&
operator << ( ostream& out, const Function& fn )
{
	out<<"(\""<<fn.fname<<"\" \""<<fn.ename<<"\" "<<( fn.isc ? "#t" : "#f" )
	   <<' '<<( fn.isp ? "#t" : "#f" )<<" (";

	for ( size_t i = 0; i < fn.args.size(); ++i )
	{
		if ( i )
			out<<' ';
		out<<fn.args[i];
	}

	return out<<"))";
}
