/*
 * Frost - use C++ multi methods as if they were a native feature
 * Copyright (C) 1999, 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_hierarchy_hh
#define _frost_hierarchy_hh


// includes

#include <iostream>
#include <map>
#include <set>
#include <string>
#include <vector>
#include "function.hh"
#include "tablegrp.hh"

using namespace std;


class Class;
class Table;
class VTable;
class vti_base;


/**	Manages hierarchies of classes.
	\author Julian v. Bock
	\see Class */
class Hierarchy
{
public:
	/** default constructor */
	Hierarchy();
	
	/** destructor */
	~Hierarchy();

	/** creates a new class that is part of the hierarchy
		\param name class name
		\return new class */
	Class&
	AddClass( const string& name );

	/** finds a class in the hierarchy
		\param n resolved class name
		\return class */
	Class&
	GetClass( const string& n );

	/** queries for a class in the hierarchy
		\param n resolved class name
		\return pointer to class or 0 if the class was not found */
	Class*
	QueryClass( const string& n );

	/** registers a function
		\param fn function */
	void
	AddFunc( Function::ptr fn );

	/** adds a virtual table
		\param n virtual table name
		\param o owner class
		\param s table size
		\return new virtual table */
	VTable&
	AddVTable( const string& n, Class& o, size_t s );

	/** \return true if there is at least one multi method */
	bool
	HasMM() const;

	/** builds multi methods and dumps the assembly code to an output stream
		\param out output stream
		\return true if all multi methods were built successfully */
	bool
	DumpMMs( ostream& out );

	/** reads a hierarchy representation from a stream
		\param in input stream */
	void
	Read( istream& in );

	/** reads symbol information from elf files, completes virtual tables,
		and adds functions.
		\param f file name */
	void
	ReadELF( const string& f );

	/** \return frost table group */
	TableGrp&
	TableGroup() { return tgrp; }

	/** output stream operator
		\param out output stream
		\param h hierarchy */
	friend ostream&
	operator << ( ostream& out, const Hierarchy& h );
	
private:
	typedef map< string, Class* > class_map;
	typedef map< string, VTable* > vtbl_map;
	typedef map< string, Table* > ftbl_map;
	typedef set< Function::ptr > func_set;
	typedef set< string > fn_set;
	typedef vector< Class* > class_vec;

	/// classes
	class_map	classes;
	/// frost tables (only used for imports)
	ftbl_map	ftbls;
	/// virtual tables
	vtbl_map	vtbls;
	/// classes in definition order
	class_vec	dcl;
	/// functions
	func_set	funcs;
	/// encoded function names
	fn_set		fnames;
	/// frost table group
	TableGrp	tgrp;

	/** builds and process the function list which was produced by one of
		the read_elf_* methods.
		\param f file name
		\param fns function name set */
	void
	process_flist( const string& f, fn_set& fns );
	
	/** reads a class representation
		\param in input stream */
	void
	read_class( istream& in );

	/** reads frost table information
		\param in input stream
		\param cl class being processed */
	void
	read_ftable( istream& in, Class& cl );

	/** reads virtual table / base class information
		\param in input stream
		\param cl class being processed */
	void
	read_base( istream& in, Class& cl );

	/** reads virtual base information
		\param in input stream
		\param cl class being processed */
	void
	read_vbase( istream& in, Class& cl );

	/** reads frost table hierarchy information
		\param in input stream */
	void
	read_fthier( istream& in );

	/** reads frost table hierarchy information
		\param in input stream
		\param p parent frost table */
	void
	read_fthier_p( istream& in, Table& p );

	/** reads a virtual table
		\param in input stream */
	void
	read_vtable( istream& in );

	/** reads a virtual table slot
		\param in input stream
		\param tbl virtual / typeinfo table */
	void
	read_vti_slot( istream& in, vti_base& tbl );

	/** reads a virtual table delta value
		\param in input stream
		\param tbl virtual / typeinfo table */
	void
	read_vti_delta( istream& in, vti_base& tbl );

	/** reads function information
		\param in input stream */
	void
	read_function( istream& in );

	/** reads function argument information
		\param in input stream
		\param fn function being processed */
	void
	read_fn_arg( istream& in, Function::ptr fn );

	/** completes virtual tables and functions which are stored in an object
		file
		\param f file name */
	void
	read_elf_o( const string& f );

	/** completes virtual tables and functions which are stored in a static
		library
		\param f file name */
	void
	read_elf_a( const string& f );

	/** completes virtual tables and functions which are stored in a shared
		library
		\param f file name */
	void
	read_elf_so( const string& f );
};

#endif // _frost_hierarchy_hh
