/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/stl_util.hh>
#include <uds/str_util.hh>
#include <cctype>
#include "class.hh"
#include "disp.hh"
#include "dthunk.hh"
#include "function.hh"
#include "mm.hh"
#include "progb.hh"
#include "table.hh"

using namespace uds;


// MMethod implementation

namespace
{

void
strip_frost_sym( string& dn )
{
	size_t fso = dn.find( SYM_PREFIX );
	size_t cso = fso;

	while ( isdigit( dn[--cso] ) );
	++cso;

	size_t cnt_len = fso - cso;
	size_t cnt = VarConv< size_t >( dn.substr( cso, cnt_len ) );

	fso += strlen( SYM_PREFIX );
	while ( dn[fso++] != '_' );

	size_t len = fso - cso;
	dn.replace( cso, len, StringVar( cnt - len + cnt_len ) );
}

}

MMethod::MMethod( const Function& f, TableGrp& tg ) :
	tgrp( tg )
{
	_frost_Log( LOG_DEBUG + 1, "fwva group created for "
				+ f.DemName() );
	mf_set.insert( &f );

	root = &f;

	static int uid_c = 0;
	uid = uid_c++;
}

MMethod::~MMethod()
{
}

string
MMethod::Name() const
{
	return root->DemName( true );
}

bool
MMethod::AddMemberMaybe( const Function& f )
{
	_frost_Log( LOG_DEBUG + 2, "AddMemberMaybe: " + root->DemName()
				+ " - " + f.DemName() );

	// we shouldn't try this if the function names do not match
	
	if ( root->FName() != f.FName() )
		_uds_abort;

	int c = f.MMG_Compare( *root );
	if ( c < 0 )
		return false;

	mf_set.insert( &f );

	if ( c )
	{
		_frost_Log( LOG_DEBUG + 2, "got new root" );
		root = &f;
	}
	else
		_frost_Log( LOG_DEBUG + 2, "added" );
	
	return true;
}

bool
MMethod::BuildMM( disp_vec& disps ) const
{
	const Function::type_vec& rtv = root->Args();
	ts_vec tree;

	for ( size_t i = 0; i < rtv.size(); ++i )
	{
		if ( rtv[i].VArg() )
		{
			tree.push_back( tab_set() );
			rtv[i].GetClass()->DTableSet( tree.back() );
		}
	}

	if ( tree.empty() )
		_uds_abort;

	for ( size_t i = 0; i < tree.size(); ++i )
		for ( ts_iter j = tree[i].begin(); j != tree[i].end(); ++j )
			( *j )->AddFragment( uid, i );

	// add global dispatchers

	so_vec soffs = root->VASOffsVec();
	cl_vec rcls = root->VAClassVec();
	size_t fof = rcls[0]->MainFTable().FrostVTableOffset();
	set< string > gdisps;
	
	for ( mf_set_t::const_iterator i = mf_set.begin();
		  i != mf_set.end(); ++i )
	{
		string dn = ( *i )->EName();
		strip_frost_sym( dn );

		_frost_Log( LOG_DEBUG + 1, "global dispatcher " + dn );

		Dispatcher* disp = new Dispatcher( dn, tgrp, uid, 0, 0, soffs[0], fof,
										   true );
		
		cl_vec fn_cls = ( *i )->VAClassVec();
		for ( size_t j = 0; j < fn_cls.size(); ++j )
		{
			size_t vof = fn_cls[j]->TableOffset( *rcls[j] );
			
			if ( vof )
				disp->Adjust( soffs[j], vof );
		}

		disps.push_back( disp );
	}

	bool succ = build_mm( disps, tab_vec(), tree, root->VASOffsVec(), rcls,
						  "", 0, true );

	return succ;
}

bool
MMethod::build_mm( disp_vec& disps, const tab_vec& fixt, const ts_vec& tree,
				   const so_vec& soffs, const cl_vec& rcls, const string& dnt,
				   int lev, bool succ ) const
{
	tab_vec nfixt( fixt );
	nfixt.push_back( 0 );

	size_t co = 0;
	for ( tab_set::iterator i = tree[lev].begin(); i != tree[lev].end(); ++i )
	{
		Table& ctab = **i;
		Class& ccl = ctab.GetClass();
		nfixt.back() = &ctab;
		string ndnt = dnt + '_' + ccl.EName();

		if ( lev < static_cast< int >( tree.size() ) - 1 )
		{
			string dname = root->EName() + '_' + StringVar( lev + 1 ) + ndnt;

			disps.push_back( new Dispatcher(
								 dname, tgrp, uid, lev + 1,
								 ( ctab.FragmentSize( uid, lev ) + co )
								 * sizeof( void* ), soffs[lev + 1],
								 ctab.FrostVTableOffset(), false ) );

			ctab.AddFragMember( uid, lev, dname );

			succ &= build_mm( disps, nfixt, tree, soffs, rcls, ndnt,
							  lev + 1, succ );
		}
		else
		{
			string err;
			m_ptr fn = resolve_call( nfixt, err );

			if ( !fn )
			{
				if ( succ )
				{
					Self().PError( "error while building fwva group "
								   + Name() + "\ncall is ambigous for the "
								   "following sets of classes:" );
				}

				Self().PError( err );
				succ = false;
			}
			else
			{
				if ( fn->IsPure() )
					ctab.AddFragMember( uid, lev, "__cxa_pure_virtual" );
				else
				{
					DispThunk* dt = new DispThunk( root->EName() + "thn"
												   + ndnt, fn->EName() );
				
					for ( size_t i = 0; i < rcls.size(); ++i )
					{
						int vof = nfixt[i]->GetClass().TableOffset( *rcls[i] );

						if ( vof )
							dt->Adjust( soffs[i], -vof );
					}

					if ( dt->Empty() )
					{
						delete dt;
						ctab.AddFragMember( uid, lev, fn->EName() );
					}
					else
					{
						disps.push_back( dt );
						ctab.AddFragMember( uid, lev, dt->Name() );
					}
				}
			}
		}

		++co;
	}

	return succ;
}

MMethod::m_ptr
MMethod::resolve_call( const tab_vec& tv, string& err ) const
{
	typedef vector< Class* > cl_vec;
	typedef map< m_ptr, cl_vec > cand_map;

	cand_map	cands;
	// remaining candidates
	mf_set_t	rcands( mf_set );
	// 'exact match' flag

	// setup candidate map and erase all function that do not match

	for ( mf_set_t::const_iterator i = mf_set.begin(); i != mf_set.end(); ++i )
	{
		cl_vec cv = ( *i )->VAClassVec();
		if ( cv.size() != tv.size() )
			_uds_abort;

		size_t j = 0;
		for ( ; j < tv.size(); ++j )
		{
			int m;
			if ( !tv[j]->GetClass().Match( *cv[j], m ) || m < 0 )
				break;
		}

		if ( j == tv.size() )
			cands[*i] = cv;
		else
			rcands.erase( *i );
	}
	
	for ( size_t i = 0; i < tv.size(); ++i )
	{
		Class& arg = tv[i]->GetClass();
		int bm = 0x7fffffff;
		mf_set_t mfs;

		for ( cand_map::iterator c = cands.begin(); c != cands.end(); ++c )
		{
			Class& cand = *c->second[i];

			int m;
			if ( !arg.Match( cand, m ) || m < 0 || m > bm )
				rcands.erase( c->first );
			else
			{
				if ( m < bm )
				{
					for ( mf_set_t::iterator i = mfs.begin(); i != mfs.end();
						  ++i )
						rcands.erase( *i );

					mfs.clear();
					bm = m;
				}
				
				mfs.insert( c->first );
			}
		}
	}
	
	if ( rcands.size() == 1 )
		return *rcands.begin();

	for ( size_t i = 0; i < tv.size(); ++i )
	{
		if ( i )
			err += ", ";

		err += tv[i]->GetClass().Name();
	}

	return 0;
}
