/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/exit.hh>
#include <uds/procstream.hh>
#include <uds/io_util.hh>
#include <uds/stl_util.hh>
#include <uds/str_util.hh>
#include <uds/sys_util.hh>
#include <bfd.h>
#include <fcntl.h>
#include <map>
#include <sstream>
#include <unistd.h>
#include "exception.hh"
#include "prog.hh"
#include "config.h"

using namespace std;
using namespace uds;


namespace
{

// utility funcs

string
mk_version_string()
{
	stringstream s;
	s<<"frost "<<VERSION<<" ("<<VERSION_TAG<<")";
	return s.str();
}


// command line table

class CmdL
{
public:
	enum
	{
		help = 1,
		dump_ffo,
		keep_asm,
		keep_cl,
		keep_fl,
		lcs_as,
		lcs_cpl,
		lcs_pp,
		lib,
		libpath,
		logfile,
		outfile,
		pdelim,
		plog,
		save_tmp,
		spec_cpp,
		spec_cpp2,
		static_link,
		twoarg,
		version
	};
	
	CmdL();
	
	typedef unsigned char value_t;
	
	value_t
	operator () ( const string& k ) const
		{ return SafeDeref( MapQueryC( table, k ) ); }
	
private:
	map< string, value_t > table;
};

CmdL::CmdL()
{
	table["--"]				= pdelim;
	table["-c"]				= lcs_as;
	table["-E"]				= lcs_pp;
	table["-Fka"]			= keep_asm;
	table["-Fkc"]			= keep_cl;
	table["-Fkf"]			= keep_fl;
	table["-Fko"]			= dump_ffo;
	table["-h"]				= help;
	table["--help"]			= help;
	table["-I"]				= twoarg;
	table["-L"]				= libpath;
	table["-l"]				= lib;
	table["--logfile"]		= logfile;
	table["-M"]				= spec_cpp;
	table["-MM"]			= spec_cpp;
	table["-MF"]			= spec_cpp2;
	table["-MG"]			= spec_cpp;
	table["-MP"]			= spec_cpp;
	table["-MT"]			= spec_cpp2;
	table["-MQ"]			= spec_cpp2;
	table["-MD"]			= spec_cpp;
	table["-MMD"]			= spec_cpp;
	table["-o"]				= outfile;
	table["--plog"]			= plog;
	table["-save-temps"]	= save_tmp;
	table["-S"]				= lcs_cpl;
	table["-static"]		= static_link;
	table["--version"]		= version;
}

}


// ProgState implementation

ProgState::ProgState()
{
}

ProgState::~ProgState()
{
}

ProgState&
ProgState::Self()
{
	static ProgramState s;
	return s;
}


// ProgramState implementation

void
ProgramState::CmdLine( const vector< string >& args )
{
	// default logfile name
	string logf = "frost.log";
	// true if '--' was specified on the command line
	bool pdelim = false;
	// gcc args
	string_vec gcc_args;
	// libraries
	string_vec libs;

	// add libstdc++
	libs.push_back( "stdc++" );

	// add compiler specific library paths
	{
		stringstream ss( LIB_PATHS );
		string lp;
		
		while ( true )
		{
			ss>>lp;

			if ( !ss )
				break;

			libpaths.push_back( lp );
		}
	}

	// add linker default library paths
	{
		ProcStream in( "ld --verbose | grep SEARCH_DIR | "
					   "sed -e 's/SEARCH_DIR(//g' -e 's/);//g' -e 's/\"//g'" );

		if ( !in )
			_uds_abort;

		while ( true )
		{
			string lp;
			in>>lp;

			if ( !in )
				break;

			libpaths.push_back( lp );
		}
	}
	
	for ( size_t i = 0; i < args.size(); ++i )
	{
		static CmdL cmdl;

		if ( i + 1 < args.size() )
		{
			switch ( cmdl( args[i] ) )
			{
			case CmdL::lib:
				libs.push_back( args[++i] );
				continue;
				
			case CmdL::libpath:
				libpaths.push_back( args[++i] );
				continue;
				
			case CmdL::logfile:
				logf = args[++i];
				continue;

			case CmdL::outfile:
				out_file = args[++i];
				continue;

			case CmdL::plog:
				log.PLogLevel( VarConv< int >( args[++i] ) );
				continue;

			case CmdL::spec_cpp2:
				cpp_opts = cpp_opts + args[i] + ' ' + args[i + 1] + ' ';
				++i;
				continue;

			case CmdL::twoarg:
				gcc_args.push_back( args[i] );
				gcc_args.push_back( args[++i] );
				continue;

			default:
				break;
			}
		}

		switch ( cmdl( args[i] ) )
		{
		case CmdL::help:
			PrintOptionSummary();
			Exit( 0 );

		case CmdL::dump_ffo:
			flags |= dump_ffo;
			continue;

		case CmdL::keep_asm:
			flags |= keep_asm;
			continue;

		case CmdL::keep_cl:
			flags |= keep_cl;
			continue;

		case CmdL::keep_fl:
			flags |= keep_fl;
			continue;

		case CmdL::lcs_as:
			lcs = assemble;
			continue;

		case CmdL::lcs_cpl:
			lcs = compile;
			continue;

		case CmdL::lcs_pp:
			lcs = preprocess;
			continue;

		case CmdL::pdelim:
			pdelim = true;
			continue;

		case CmdL::save_tmp:
			flags |= save_tmp;
			continue;

		case CmdL::spec_cpp:
			cpp_opts = cpp_opts + args[i] + ' ';
			continue;

		case CmdL::static_link:
			flags |= st_link;
			break;
			
		case CmdL::version:
			cout<<Version()<<endl;
			Exit( 0 );

		default:
			break;
		}

		if ( args[i][0] == '-' && !pdelim )
		{
			if ( args[i].size() > 2 )
			{
				if ( args[i][1] == 'L' )
				{
					libpaths.push_back( args[i].substr( 2 ) );
					continue;
				}
				else if ( args[i][1] == 'l' )
				{
					libs.push_back( args[i].substr( 2 ) );
					continue;
				}
			}
			
			gcc_args.push_back( args[i] );
		}
		else
		{
			// determine start compilation stage for file
			cstage_t ss = filename_to_cstage( args[i] );
			ifq[ss].push( args[i] );
		}
	}

	for ( size_t i = 0; i < libs.size(); ++i )
		ifq[link].push( LibraryName( libs[i] ) );

	gcc_cmd_nd = gcc_cmd = CXX " ";
	for ( size_t i = 0; i < gcc_args.size(); ++i )
	{
		gcc_cmd += gcc_args[i] + ' ';

		if ( gcc_args[i].size() > 1 && gcc_args[i][1] != 'g' )
			gcc_cmd_nd += gcc_args[i] + ' ';
	}

	string lid = "frost";
	if ( log.PLogLevel() )
		log.OpenLog( lid, logf.c_str(),
					 Logger::use_ident | Logger::log_pid | Logger::add_tabs
					 | Logger::print_pri | Logger::print_time );
}
	
void
ProgramState::Execute()
{
	try
	{
		_frost_Log( LOG_NOTICE, Version() + " started\n" );

		SetSigHandlers();
		check_cxx_version();
		bfd_init();
		ProcInputFiles();

		log( LOG_NOTICE, "terminating (status: success)" );
	}
	catch ( error& e )
	{
		log( LOG_ALERT, "ERROR: " + e.Info() );
		log( LOG_ALERT, "terminating (status: error)" );
		
		throw;
	}
	catch ( Exception& e )
	{
		log( LOG_ALERT, "fatal exception\n" + e.Message() );
		log( LOG_ALERT, "terminating (status: error)" );
		
		throw;
	}
	catch ( Exit_t& e )
	{
		if ( log( LOG_NOTICE ) )
		{
			stringstream s;
			s<<"terminating (status: "<<e.ReturnValue()<<", "
			 <<( e.ReturnValue() ? "error" : "ok" )<<')';
			log( LOG_NOTICE, s.str() );
		}
		
		throw;
	}
	catch ( ... )
	{
		log( LOG_ALERT, "terminating (status: error)" );
		
		throw;
	}
}

string
ProgramState::TmpFileName()
{
	string file = uds::TmpFileName();
	tfiles.push_back( file );

	return file;
}

void
ProgramState::PError( const string& m )
{
	cerr<<m<<endl;

	log( LOG_ERR, m );
}

string
ProgramState::LibraryName( const string& n ) const
{
	string lh = "/lib" + n;
	string_vec suf;
	
	if ( ~flags & st_link )
		suf.push_back( ".so" );
	suf.push_back( ".a" );

	for ( size_t i = 0; i < suf.size(); ++i )
	{
		for ( size_t j = 0; j < libpaths.size(); ++j )
		{
			string lib = libpaths[j] + lh + suf[i];
			int fd = open( lib.c_str(), O_RDONLY );

			if ( fd >= 0 )
			{
				close( fd );
				return lib;
			}
		}
	}

	throw UserError( "cannot find -l" + n );
}

string
ProgramState::RepSuffix( const string& fn, const string& ns, bool s ) const
{
	string sp = s ? fn.substr( fn.rfind( "/" ) + 1, string::npos ) : fn;
	return sp.substr( 0, sp.rfind( "." ) ) + ns;
}

ProgramState::ProgramState() :
	log( 0, 0 ), ifq( 4 ), flags( 0 ), lcs( link )
{
}

ProgramState::~ProgramState()
{
	log( LOG_DEBUG, "deleting temporary files" );
	
	for ( size_t i = 0; i < tfiles.size(); ++i )
		unlink( tfiles[i].c_str() );
}

void
ProgramState::PrintOptionSummary()
{
	cout<<Version()<<endl<<
		"Copyright (C) 1999 - 2002  Julian v. Bock\n"
		"This program is licensed under the GPL\n"
		"\nfrost options:\n"
		"  -h, --help            this message.\n"
		"  -Fka                  keep frost assembler file.\n"
		"  -Fkc                  keep class files.\n"
		"  -Fkf                  keep function lists.\n"
		"  -Fko                  dump final frost object file\n."
		"  --logfile f           use logfile [f].\n"
		"  --plog l              set log level to [l]. default is 0.\n"
		"  -save-temps           save the usual intermediate files.\n"
		"  --version             print the version number and exit.\n";
}

const string&
ProgramState::Version()
{
	static const string v = mk_version_string();
	return v;
}

void
ProgramState::check_cxx_version() const
{
	ProcStream ps( CXX " --version" );
	string cxxv = ReadLine( ps );

	if ( cxxv != UDS_CXX_VERSION )
		throw error( "bad compiler version (" + cxxv + ")\n"
					 "frost was configured for version " UDS_CXX_VERSION );
}

ProgramState::cstage_t
ProgramState::filename_to_cstage( const string& fn ) const
{
	size_t pos = fn.rfind( "." );
	if ( pos != string::npos )
	{
		string suf = fn.substr( pos + 1, fn.size() - pos - 1 );

		if ( suf == "cc"
			 || suf == "cxx"
			 || suf == "cpp"
			 || suf == "C" )
			return preprocess;

		if ( suf == "ii" )
			return compile;

		if ( suf == "s" )
			return assemble;
	}

	return link;
}

string
ProgramState::cstage_output_filename( const string& fn, cstage_t cs )
{
	// check whether the last stage is reached
	if ( cs == lcs || flags & save_tmp )
	{
		if ( cs == lcs && !out_file.empty() )
			return out_file;

		switch ( cs )
		{
		case preprocess:
			return RepSuffix( fn, ".ii", true );
			
		case compile:
			return RepSuffix( fn, ".s", true );

		case assemble:
			return RepSuffix( fn, ".o", true );

		case link:
			return "a.out";

		default:
			_uds_abort;
		}
	}

	if ( cs >= link )
		_uds_abort;

	return TmpFileName();
}

string
ProgramState::fo_filename( const string& fn )
{
	if ( lcs == compile || lcs == assemble || flags & save_tmp )
		return RepSuffix( fn, ".fo", true );

	return TmpFileName();
}

void
ProgramState::exec_gcc( const string& cmd, bool nd )
{
	string gc = ( nd ? gcc_cmd_nd : gcc_cmd ) + cmd;
	log( LOG_INFO, "executing " + gc );

	try
	{
		System( gc.c_str() );
	}
	catch ( child_error& e )
	{
		Exit( 1 );
	}
}
