/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_prog_hh
#define _frost_prog_hh


// includes

#include <uds/streamsocket.hh>
#include <queue>
#include "progb.hh"

using namespace std;
using namespace uds;


class Hierarchy;


/** holds information about command line options and things to be done
	\author Julian v. Bock */
class ProgramState : public ProgState
{
public:
	/** processes command line arguments
		\param arg argument
		\arg throws uds::bad_cmd_line_opt */
	virtual void
	CmdLine( const vector< string >& args );
	
	/** does everything ;-) */
	virtual void
	Execute();
	
	/** checks whethers flags are set
		\param f flags */
	virtual bool
	Flags( flag_t f ) const { return ( flags & f ) == f; }
	
	/** \return logging facility */
	virtual const Logger&
	Log() { return log; }

	/** generates a temporary file name. temporary files are deleted upon
		program termination
		\return temporary file name */
	virtual string
	TmpFileName();

	/** prints an error message to stderr (and to the logfile)
		\param m error message */
	virtual void
	PError( const string& m );

	/** determines the absolute filename of a library
		\param n library name
		\return absolute library name */
	virtual string
	LibraryName( const string& n ) const;

	/** replaces the suffix of a file
		\param fn file name
		\param ns new file suffix (including dot)
		\param s 'strip leading path' flag */
	virtual string
	RepSuffix( const string& fn, const string& ns, bool s ) const;
	
private:
	/// compile stage values
	enum
	{
		preprocess = 0,	///< preprocess stage
		compile = 1,	///< compile stage
		assemble = 2,	///< assemble stage
		link = 3		///< link stage
	};

	typedef vector< string > string_vec;
	typedef unsigned char cstage_t;
	typedef queue< string > fileq_t;
	typedef vector< fileq_t > ifq_vec;

	/// logging facility
	Logger			log;
	/// temporary files
	string_vec		tfiles;
	/// library paths
	string_vec		libpaths;
	/// input file queues
	ifq_vec			ifq;
	/// frost object files
	string_vec		fofq;
	/// linker input files
	string_vec		linkq;
	/// elf files that need to be scanned for vtables and functions
	string_vec		relfq;
	/// output file
	string			out_file;
	/// gcc command line
	string			gcc_cmd;
	/// gcc command line without -g* options
	string			gcc_cmd_nd;
	/// special preprocessor command line options
	string			cpp_opts;
	/// flags
	flag_t			flags;
	/// last compile stage
	cstage_t		lcs;
	
	/** default constructor */
	ProgramState();
	
	/** destructor */
	virtual
	~ProgramState();
	
	/** prints a short option summary to stdout */
	static void
	PrintOptionSummary();
	
	/** \return version string */
	static const string&
	Version();
	
	/** sets signal handlers */
	void
	SetSigHandlers() const;

	/** processes input files */
	void
	ProcInputFiles();

	/** checks the c++ compiler version */
	void
	check_cxx_version() const;

	/** determines the compilation stage for a file name
		\param fn file name
		\return compilation stage */
	cstage_t
	filename_to_cstage( const string& fn ) const;

	/** determines the output file name
		\param fn file name
		\param cs current compile stage
		\return output filename */
	string
	cstage_output_filename( const string& fn, cstage_t cs );

	/** determines frost object file name
		\param fn input file name
		\return freost object file name */
	string
	fo_filename( const string& fn );

	/** executes gcc
		\param cmd command
		\param nd 'no debug options' flag */
	void
	exec_gcc( const string& cmd, bool nd = false );

	/** read a frost object file
		\param hier hierarchy object
		\param file frost object file name */
	void
	read_fobj( Hierarchy& hier, const string& file );

	/** processes one file
		\param file file
		\param ss start stage */
	void
	proc_file( const string& file, cstage_t ss );

	/** links files */
	void
	link_files();
	
	friend class ProgState;
};

#endif // _frost_prog_hh
