/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_progb_hh
#define _frost_progb_hh


// includes

#include <uds/log.hh>
#include <uds/stdexcpt.hh>
#include <vector>

using namespace std;
using namespace uds;


class ProgramState;


/** base class for ProgramState class. It defines only the visible interface
	to reduce compilation dependencies and compilation time.
	\author Julian v. Bock */
class ProgState
{
public:
	/// 'flag' type
	typedef unsigned char flag_t;
	/// flag values
	enum
	{
		st_link		= 0x1,	///< link against static libraries
		dyn_fol		= 0x2,	///< use dynamic frost object loader
		keep_asm	= 0x4,	///< keep frost assembler file
		keep_cl		= 0x8,	///< keep class files
		keep_fl		= 0x10,	///< keep function lists
		dump_ffo	= 0x20,	///< dump final frost object
		save_tmp	= 0x40	///< save temporary files
	};
	
	/** returns the one and only ProgramState instance
		\return instance */
	static ProgState&
	Self();
	
	/** processes command line arguments
		\param arg argument
		\arg throws uds::bad_cmd_line_opt */
	virtual void
	CmdLine( const vector< string >& args ) = 0;
	
	/** does everything ;-) */
	virtual void
	Execute() = 0;
	
	/** checks whethers flags are set
		\param f flags */
	virtual bool
	Flags( flag_t f ) const = 0;
	
	/** \return logging facility */
	virtual const Logger&
	Log() = 0;

	/** generates a temporary file name. temporary files are deleted upon
		program termination
		\return temporary file name */
	virtual string
	TmpFileName() = 0;

	/** prints an error message to stderr and the logfile
		\param m error message */
	virtual void
	PError( const string& m ) = 0;

	/** determines the absolute filename of a library
		\param n library name
		\return absolute library name */
	virtual string
	LibraryName( const string& n ) const = 0;

	/** replaces the suffix of a file
		\param fn file name
		\param ns new file suffix (including dot)
		\param s 'strip leading path' flag */
	virtual string
	RepSuffix( const string& fn, const string& ns, bool s ) const = 0;
	
private:
	/** default constructor */
	ProgState();
	
	/** destructor */
	virtual
	~ProgState();
	
	friend class ProgramState;
};


inline ProgState&
Self()
{
	return ProgState::Self();
}

inline void
Log( Logger::loglev_t p, const string& m )
{
	Self().Log()( p, m );
}

inline bool
Log( Logger::loglev_t p )
{
	return Self().Log()( p );
}

#define _frost_Log( lev, msg )                                                \
	do                                                                        \
	{                                                                         \
		if ( ::Log( lev ) )                                                   \
			::Log( lev, msg );                                                \
	} while ( false )

#endif // _frost_progb_hh
