/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */


// includes

#include <uds/stdexcpt.hh>
#include <uds/stl_util.hh>
#include <uds/str_util.hh>
#include "ascode.hh"
#include "class.hh"
#include "progb.hh"
#include "sio.hh"
#include "table.hh"
#include "tablegrp.hh"
#include "titable.hh"
#include "vtable.hh"
#include "config.h"

using namespace uds;


// Table implementation

Table::Table( const string& n, TableGrp& tg, Class& ocl, int fvo ) :
	name( n ), tgrp( tg ), cl( ocl ), fv_offs( fvo )
{
	_frost_Log( LOG_DEBUG + 2, "new frost table " + name + " fvo "
				+ StringVar( fvo ) );

	tgrp.AddTrunk( *this );
}

Table::~Table()
{
	tgrp.RemoveTrunk( *this );
}

void
Table::SetParent( Table& tbl )
{
	if ( this == &tbl )
		return;

	_frost_Log( LOG_DEBUG + 2, "ftable parent: " + name + " -> " + tbl.name );

	tgrp.RemoveTrunk( *this );
	tbl.leaves.insert( this );
}

void
Table::DTableSet( tbl_set& ts )
{
	ts.insert( this );

	for ( tbl_set::iterator i = leaves.begin(); i != leaves.end(); ++i )
		( *i )->DTableSet( ts );
}

void
Table::AdjustFrostVTableOffset( int dt )
{
	_frost_Log( LOG_DEBUG + 2, "fvo adjust: table " + name + " delta "
				+ StringVar( dt ) );
	
	if ( dt + static_cast< int >( fv_offs ) < -4 )
		_uds_abort;

	fv_offs += dt;
}

void
Table::AddFragment( int s, int l )
{
	_frost_Log( LOG_DEBUG + 2, "new fragment: table " + name
				+ " section " + StringVar( s )
				+ " level " + StringVar( l ) );
	
	MapInsert< internal_error >( frags, make_pair( s, l ),
								 vector< string >() );
}

void
Table::AddFragMember( int s, int l, const string& d )
{
	_frost_Log( LOG_DEBUG + 2, "new frag member: table " + name
				+ " section " + StringVar( s )
				+ " level " + StringVar( l )
				+ " name " + d );
	
	vector< string >& frag =
		MapFind< internal_error >( frags, make_pair( s, l ) );

	_frost_Log( LOG_DEBUG + 2, "assigned offset " + StringVar( frag.size() ));
	
	frag.push_back( d );
}

size_t
Table::FragmentSize( int s, int l ) const
{
	const vector< string >& frag =
		MapFindC< internal_error >( frags, make_pair( s, l ) );

	return frag.size();
}

void
Table::DumpHierarchy( ostream& out ) const
{
	out<<"\""<<name<<"\" (";

	bool f = true;
	for ( tbl_set::const_iterator i = leaves.begin(); i != leaves.end(); ++i )
	{
		if ( f )
			f = false;
		else
			out<<' ';
		
		( *i )->DumpHierarchy( out );
	}

	out<<')';
}

void
Table::Arrange( int o )
{
	for ( frag_map::const_iterator i = frags.begin(); i != frags.end(); ++i )
	{
		frag_key sl = i->first;

		if ( tgrp.HasFragOffset( sl ) )
		{
			o -= i->second.size();
			tgrp.SetFragOffset( sl, o * sizeof( void* ) );
			ext_table( sl );
		}
	}
	
	for ( tbl_set::iterator i = leaves.begin(); i != leaves.end(); ++i )
		( *i )->Arrange( o );
}

void
Table::InsertOc( set< Class* >& cs )
{
	if ( !frags.empty() )
		cs.insert( &cl );

	for ( tbl_set::iterator i = leaves.begin(); i != leaves.end(); ++i )
		( *i )->InsertOc( cs );
}

void
Table::EmitCode( ostream& out ) const
{
	if ( frags.empty() )
		return;
	
	size_t sz = 0;

	for ( frag_map::const_iterator i = frags.begin(); i != frags.end(); ++i )
		sz += i->second.size();

	DefineFTableHead( out, name, sz + 2 );

	// the fragments have negative offsets; print them in reverse order

	for ( int i = fkeys.size() - 1; i >= 0; --i )
	{
		const vector< string >& frag
			= MapFindC< internal_error >( frags, fkeys[i] );

		for ( size_t j = 0; j < frag.size(); ++j )
			PrintTableEntry( out, frag[j] );
	}

	// print the contents of the original typeinfo table

	DefineFTableDef( out, name );
	cl.Vtbl().TInfo().EmitCode( out );

	// print the code of the leave tables
	
	for ( tbl_set::const_iterator i = leaves.begin(); i != leaves.end(); ++i )
		( *i )->EmitCode( out );
}

void
Table::ext_table( const frag_key& sl )
{
	fkeys.push_back( sl );
	
	for ( tbl_set::iterator i = leaves.begin(); i != leaves.end(); ++i )
		( *i )->ext_table( sl );
}


// misc. implementation

ostream&
operator << ( ostream& out, const Table& t )
{
	return out<<"(\""<<t.name<<"\" "<<t.fv_offs<<')';
}
