/*
 * frost - C++ multi methods as if they were a native feature
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _frost_table_hh
#define _frost_table_hh


// includes

#include <iostream>
#include <map>
#include <set>
#include <string>
#include <vector>

using namespace std;


class Class;
class TableGrp;


/** frost table class
	\author Julian v. Bock */
class Table
{
public:
	typedef set< Table* > tbl_set;

	/** \param n table name
		\param tg table group
		\param ocl owner class
		\param fvo initial frost vtable offset */
	Table( const string& n, TableGrp& tg, Class& ocl, int fvo );

	/** destructor */
	~Table();

	/** \return owner class */
	Class&
	GetClass() const { return cl; }

	/** sets the parent table
		\param tbl parent table */
	void
	SetParent( Table& tbl );

	/** fills a set with this table and all derived tables
		\param ts set to be filled */
	void
	DTableSet( tbl_set& ts );

	/** \return frost vtable offset */
	int
	FrostVTableOffset() const { return fv_offs; }
	
	/** adjusts the offset of the frost table pointer in the virtual
		table
		\param dt delta */
	void
	AdjustFrostVTableOffset( int dt );

	/** adds a frost table fragment
		\param s section
		\param l level */
	void
	AddFragment( int s, int l );

	/** adds a frost table fragment member
		\param s section
		\param l level
		\param d dispatcher function name */
	void
	AddFragMember( int s, int l, const string& d );

	/** returns the current size of a fragment
		\param s section
		\param l level */
	size_t
	FragmentSize( int s, int l ) const;

	/** dumps the table hierarchy to an output stream
		\param out output stream */
	void
	DumpHierarchy( ostream& out ) const;
		
	/** arranges frost table fragments
		\param o fragment offset */
	void
	Arrange( int o );

	/** inserts the owner classes of this and all derived tables into a set.
		\param cs class set */
	void
	InsertOc( set< Class* >& cs );

	/** prints code of this and all derived frost tables to an output stream
		\param out output stream */
	void
	EmitCode( ostream& out ) const;

	/** output stream operator
		\param out output stream
		\param t table */
	friend ostream&
	operator << ( ostream& out, const Table& t );

private:
	typedef pair< int, int > frag_key;
	typedef vector< frag_key > fk_vec;
	typedef map< frag_key, vector< string > > frag_map;

	/** extends the vector of frost table fragment keys
		\param sl key */
	void
	ext_table( const frag_key& sl );

	/// table name
	string		name;
	/// table group
	TableGrp&	tgrp;
	/// owner class
	Class&		cl;
	/// offset of frost table pointer in virtual table
	int			fv_offs;
	/// derived tables
	tbl_set		leaves;
	/// frost table fragments
	frag_map	frags;
	/// fragment keys in reverse order
	fk_vec		fkeys;

	/// forbidden
	Table( const Table& );
	/// forbidden
	bool
	operator = ( const Table& );
};

#endif // _frost_table_hh
